#cython: embedsignature=True
#cython: profile=False

import os
import sys
import numpy as np
cimport numpy as np
import math
cimport cython
import pandas as pd
from libc.string cimport memcpy
np.import_array()
from time import time
from .LabeledIntervalArray_core import LabeledIntervalArray

# Set byteorder for __reduce__
byteorder = sys.byteorder


def get_include():
	"""
	Get file directory if C headers

	Parameters
	----------
		None

	Returns
	-------
		location : str
			Directory to header files
	"""

	# Grab file location
	location = os.path.split(os.path.realpath(__file__))[0]

	return location


cdef class LabeledInterval(object):
	"""
	Wrapper of C labeled_interval_t

	:class:`~aiarray.LabeledInterval` stores an interval
	
	"""

	def __init__(self, start=None, end=None, id_value=None, label=None):
		"""
		Initialize LabeledInterval

		Parameters
		----------
			start : int
				Starting position [default = None]
			end : int
				Ending position [default = None]
			id_value : int
				ID [defualt = None]
			label : str
				Label

		Returns
		-------
			None

		"""

		if start is not None:
			if id_value is None:
				id_value = 0
			# Create interval
			self.i = labeled_interval_init(start, end, id_value, 0)
			self._label = label

	# Set the interval
	cdef void set_i(LabeledInterval self, labeled_interval_t *i, str label):
		"""
		Initialize wrapper of C interval

		Parameters
		----------
			i : labeled_interval_t
				C labeled_interval_t to be wrapped

		Returns
		-------
			None
		"""

		# Set i
		self.i = i
		self._label = label


	@property
	def start(self):
		"""
		Start of interval
		"""
		return self.i.start

	@property
	def end(self):
		"""
		End of interval
		"""
		return self.i.end
	
	@property
	def id_value(self):
		"""
		ID of interval
		"""
		return self.i.id_value

	@property
	def label(self):
		"""
		Label of interval
		"""
		return self._label
	

	def __hash__(self):
		"""
		Get hash value
		"""

		return hash(repr(self))

	
	def __eq__(self, other):
		"""
		Check if there is overlap
		"""

		# Check that the classes match
		is_equal = self.__class__ == other.__class__

		# Check for overlap
		if is_equal:
			if self.label == other.label:
				if other.start < self.i.end and other.end > self.i.start:
					is_equal = True
				else:
					is_equal = False
			else:
				is_equal = False
		
		return is_equal


	def __str__(self):
		format_string = "Interval(%d-%d, %s)" % (self.start, self.end, self.label)
		return format_string


	def __repr__(self):
		format_string = "Interval(%d-%d, %s)" % (self.start, self.end, self.label)
		return format_string


	def to_list(self):
		"""
		"""

		ail = LabeledIntervalArray()
		ail.add(self.start, self.end, self.label)

		return ail


import numpy as np
cimport numpy as np
cimport cython
from libc.stdint cimport uint32_t, int32_t, int64_t
from libc.stdlib cimport malloc, free
from ailist.Interval_core cimport Interval, interval_init, interval_t
from ailist.array_query_core cimport *

cdef extern from "src/augmented_interval_list.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/ailist_add.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/ailist_construct.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/ailist_index.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/ailist_query.c":
	# C is include here so that it doesn't need to be compiled externally
	pass
	
cdef extern from "src/ailist_iter.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/ailist_coverage.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/ailist_nhits.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/ailist_wps.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/ailist_merge.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/ailist_extract.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/ailist_ops.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/ailist_filter.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/utilities.h":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/augmented_interval_list.h":

	ctypedef struct ailist_t:
		int64_t nr, mr  					# Number of regions
		interval_t *interval_list			# Regions data
		uint32_t first, last				# Record range of intervals
		int nc
		int lenC[10]
		int idxC[10]
		uint32_t *maxE
		
	ctypedef struct ailist_sorted_iter_t:
		ailist_t *ail						# Interval list
		int nc								# Number of components
		int *comp_bounds					# Label component bounds
		int *comp_used						# Components used
		interval_t *intv					# Interval
		int n								# Current position


	#-------------------------------------------------------------------------------------
	# augmented_interval_list.c
	#=====================================================================================

	# Initialize ailist_t
	ailist_t *ailist_init() nogil

	# Free ailist data
	void ailist_destroy(ailist_t *ail) nogil

	# Print AIList
	void display_list(ailist_t *ail) nogil

	# Calculate maximum length
	int ailist_max_length(ailist_t *ail) nogil

	# Calculate length distribution
	void ailist_length_distribution(ailist_t *ail, int distribution[]) nogil


	#-------------------------------------------------------------------------------------
	# ailist_add.c
	#=====================================================================================

	# Add interval to ailist_t object 
	void ailist_add(ailist_t *ail, uint32_t start, uint32_t end, uint32_t id) nogil

	# Build ailist from arrays
	void ailist_from_array(ailist_t *ail, const long starts[], const long ends[], const long ids[], int length) nogil

	# Append two ailist
	ailist_t *ailist_append(ailist_t *ail1, ailist_t *ail2) nogil


	#-------------------------------------------------------------------------------------
	# ailist_construct.c
	#=====================================================================================

	# Construct ailist: decomposition and augmentation
	void ailist_construct(ailist_t *ail, int cLen) nogil

	#-------------------------------------------------------------------------------------
	# ailist_index.c
	#=====================================================================================

	# Get intervals with id
	ailist_t *ailist_get_id(ailist_t *ail, int query_id) nogil

	# Get intervals with ids
	ailist_t *ailist_get_id_array(ailist_t *ail, const long ids[], int length) nogil


	#-------------------------------------------------------------------------------------
	# ailist_query.c
	#=====================================================================================

	# Binary search
	uint32_t binary_search(interval_t* As, uint32_t idxS, uint32_t idxE, uint32_t qe) nogil

	# Query ailist intervals
	ailist_t *ailist_query(ailist_t *ail, uint32_t qs, uint32_t qe) nogil

	# Query ailist intervals of a length
	ailist_t *ailist_query_length(ailist_t *ail, uint32_t qs, uint32_t qe, int min_length, int max_length) nogil

	# Query ailist intervals from arrays
	ailist_t *ailist_query_from_array(ailist_t *ail, const long starts[], const long ends[], int length) nogil

	# Query ailist intervals from another ailist
	ailist_t *ailist_query_from_ailist(ailist_t *ail, ailist_t *ail2) nogil

	# Query ailist interval ids from array
	array_query_t *ailist_query_id_from_array(ailist_t *ail, const long starts[], const long ends[], const long ids[], int length) nogil

	# Query ailist interval ids from another ailist
	array_query_t *ailist_query_id_from_ailist(ailist_t *ail, ailist_t *ail2) nogil
	
	
	#-------------------------------------------------------------------------------------
	# ailist_iter.c
	#=====================================================================================

	# Get component index
	int *get_comp_bounds(ailist_t *ail) nogil

	# 
	ailist_sorted_iter_t *ailist_sorted_iter_init(ailist_t *ail) nogil

	# 
	int ailist_sorted_iter_next(ailist_sorted_iter_t *iter) nogil

	# 
	void ailist_sorted_iter_destroy(ailist_sorted_iter_t *iter) nogil


	#-------------------------------------------------------------------------------------
	# ailist_coverage.c
	#=====================================================================================

	# Calculate coverage for a single interval
	void ailist_interval_coverage(ailist_t *ail, int start, int end, int coverage[]) nogil

	# Calculate coverage
	void ailist_coverage(ailist_t *ail, double coverage[]) nogil

	# Calculate coverage within bins
	void ailist_bin_coverage(ailist_t *ail, double coverage[], int bin_size) nogil

	# Calculate coverage within bins of a length
	void ailist_bin_coverage_length(ailist_t *ail, double coverage[], int bin_size, int min_length, int max_length) nogil


	#-------------------------------------------------------------------------------------
	# ailist_nhits.c
	#=====================================================================================

	# Determine number of hits for each query interval
	void ailist_nhits_from_array(ailist_t *ail, const long starts[], const long ends[],
								int length, int nhits[]) nogil

	# Determine number of hits of a length for each query interval
	void ailist_nhits_from_array_length(ailist_t *ail, const long starts[], const long ends[],
										int length, int nhits[], int min_length,
										int max_length) nogil

	# Calculate n hits within bins
	void ailist_bin_nhits(ailist_t *ail, double coverage[], int bin_size) nogil

	# Calculate n hits of a length within bins
	void ailist_bin_nhits_length(ailist_t *ail, double coverage[], int bin_size, int min_length, int max_length) nogil

	#-------------------------------------------------------------------------------------
	# ailist_wps.c
	#=====================================================================================

	# Calculate Window Protection Score
	void ailist_wps(ailist_t *ail, double wps[], uint32_t protection) nogil

	# Calculate Window Protection Score of a length
	void ailist_wps_length(ailist_t *ail, double wps[], uint32_t protection, int min_length, int max_length) nogil

	#-------------------------------------------------------------------------------------
	# ailist_merge.c
	#=====================================================================================

	# Merge nearby intervals
	ailist_t *ailist_merge(ailist_t *ail, uint32_t gap) nogil


	#-------------------------------------------------------------------------------------
	# ailist_extract.c
	#=====================================================================================

	# Extract start for ailist
	void ailist_extract_starts(ailist_t *ail, long starts[]) nogil

	# Extract end for ailist
	void ailist_extract_ends(ailist_t *ail, long ends[]) nogil

	# Extract index for ailist
	void ailist_extract_ids(ailist_t *ail, long ids[]) nogil


	#-------------------------------------------------------------------------------------
	# ailist_ops.c
	#=====================================================================================

	# Subtract intervals from region
	void subtract_intervals(ailist_t *ref_ail, ailist_t *result_ail, interval_t query_i, int j) nogil

	# Subtract two ailist_t intervals
	ailist_t *ailist_subtract(ailist_t *ref_ail, ailist_t *query_ail) nogil

	# Subtract intervals from region
	void common_intervals(ailist_t *ref_ail, ailist_t *result_ail, interval_t query_i, int j) nogil

	# Subtract two ailist_t intervals
	ailist_t *ailist_common(ailist_t *ref_ail, ailist_t *query_ail) nogil


	#-------------------------------------------------------------------------------------
	# ailist_filter.c
	#=====================================================================================

	# Filter ailist by length
	ailist_t *ailist_length_filter(ailist_t *ail, int min_length, int max_length) nogil

	# Randomly downsample
	ailist_t *ailist_downsample(ailist_t *ail, double proportion) nogil


cpdef object rebuild_AIList(bytes data, bytes b_length)


cdef class AIList(object):
	"""
	Wrapper for C ailist_t
	"""

	# AIList attributes
	cdef ailist_t *c_ailist
	cdef public bint is_constructed
	cdef public bint is_closed
	cdef public bint is_frozen

	# Methods for serialization
	cdef bytes _get_data(self)
	cdef ailist_t *_set_data(self, bytes data, bytes b_length)

	# AIList methods
	cdef void set_list(AIList self, ailist_t *input_list)
	cdef void _insert(AIList self, int start, int end, int id_value)
	cdef void _construct(AIList self, int min_length)
	cdef ailist_t *_array_id(AIList self, const long[::1] ids)
	cdef ailist_t *_interval_id(AIList self, int id_value)
	cdef ailist_t *_intersect(AIList self, int start, int end)
	cdef np.ndarray _intersect_ids(AIList self, int start, int end)
	cpdef _intersect_ids_from_array(AIList self, const long[::1] starts, const long[::1] ends, const long[::1] ids)
	cpdef _intersect_ids_from_ailist(AIList self, AIList ail)
	cdef np.ndarray _coverage(AIList self)
	cdef np.ndarray _bin_coverage(AIList self, int bin_size)
	cdef np.ndarray _bin_coverage_length(AIList self, int bin_size, int min_length, int max_length)
	cdef np.ndarray _bin_nhits(AIList self, int bin_size)
	cdef np.ndarray _bin_nhits_length(AIList self, int bin_size, int min_length, int max_length)
	cdef np.ndarray _wps(AIList self, int protection)
	cdef np.ndarray _wps_length(AIList self, int protection, int min_length, int max_length)
	cdef np.ndarray _length_dist(AIList self)
	cdef np.ndarray _nhits_from_array(AIList self, const long[::1] starts, const long[::1] ends)
	cdef np.ndarray _nhits_from_array_length(AIList self, const long[::1] starts, const long[::1] ends, int min_length, int max_length)
	cdef np.ndarray _interval_coverage(AIList self, int start, int end)


import pytest
from ailist import LabeledIntervalArray, LabeledInterval
import numpy as np
import pickle


@pytest.fixture
def interval_list():
# Test IntervalArray construction
i = LabeledIntervalArray()
i.add(10, 11, 'a')
i.add(1, 11, 'b')
i.add(15, 19, 'a')
i.add(11, 13, 'a')
i.add(19, 22, 'a')
i.add(15, 17, 'b')
i.add(20, 25, 'a')
i.add(30, 100, 'a')
i.add(1, 110, 'c')
i.add(30, 95, 'a')
i.add(20, 30, 'c')
i.add(25, 31, 'c')
i.construct()

    return i
    
    
def test_iteration(interval_list):
    # Expected results
    expected_starts = np.array([10, 1, 15, 11, 19, 15, 20, 30, 1, 30, 20, 25], dtype=int)
    
    # Test iteration
    is_Interval = 0
    for x in interval_list:
        is_Interval += isinstance(x, LabeledInterval)
    assert (is_Interval == 12)
    

def test_sorted_iteration(interval_list):
    # Expected results
    expected_sorted_starts = np.array([10, 11, 15, 19, 20, 30, 30, 1, 15, 1, 20, 25], dtype=int)
    
    # Test iteration
    is_Interval = 0
    for x in interval_list.iter_sorted:
        is_Interval += isinstance(x, LabeledInterval)
    assert (is_Interval == 12)
    

def test_nhits(interval_list):
    # Expected results
    expected_bin_nhits = np.array([2,2,2,0,2,2,2,2,2,2,2,2,2,2,2,2,2,1,
                                   1,1,1,1,
                                   1,1,1,1,2,3,2,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1], dtype=np.double)
    expected_bin_starts = np.array([10,15,20,25,30,35,40,45,50,55,60,65,70,75,80,85,90,95,
                                    0,5,10,15,
                                    0,5,10,15,20,25,30,35,40,45,50,55,60,65,70,75,80,85,90,95,100,105])
    expected_bin_nhits_length = np.array([1,2,2,0,2,2,2,2,2,2,2,2,2,2,2,2,2,1], dtype=np.double)
    expected_nhits = np.array([1, 4, 4, 2, 2, 2, 2, 2, 2, 2])

    # Test nhits
    interval_list.nhits(3, 15, 'a')

    # Test bin nhits
    bh_bins, bh = interval_list.bin_nhits(5)
    assert (bh.values == expected_bin_nhits).all()

    # Test bin nhits length
    bhl = interval_list.bin_nhits(5, 2, 100)
    assert (bhl.values == expected_bin_nhits_length).all()

    # Test nhits from arrray
    starts = np.arange(1,100,10)
    ends = starts + 10
    nhits = interval_list.nhits_from_array(starts, ends)
    assert (nhits == expected_nhits).all()


def test_bin_coverage(interval_list):
    # Expected results
    expected_bin_coverage = np.array([3,5,7,0,10,10,10,10,10,10,10,10,10,10,10,10,10,5], dtype=np.double)
    expected_bin_coverage_length = np.array([2,5,7,0,10,10,10,10,10,10,10,10,10,10,10,10,10,5], dtype=np.double)

    # Test bin coverage
    bc = interval_list.bin_coverage(5)
    assert (bc.values == expected_bin_coverage).all()

    # Test bin coverage length
    bcl = interval_list.bin_coverage(5, 2, 100)
    assert (bcl.values == expected_bin_coverage_length).all()


def test_intersection(interval_list):
    # Expected results
    expected_intersect_index = (np.array([0, 1, 1, 2, 2], dtype=np.intc), np.array([5, 0, 3, 7, 9], dtype=np.long))

    # Test intersection
    o = interval_list.intersect(3, 15, 'a')
    assert o.size == 2 and o.label_ranges['a'][0] == 10 and o.label_ranges['a'][1] == 13

    # Test intersect from array
    oi = interval_list.intersect_from_array(np.array([16,1,30]), np.array([40,15,35]), np.array(['b','a','a']))
    assert (oi[0] == expected_intersect_index[0]).all()
    assert (oi[1] == expected_intersect_index[1]).all()

    # Test intersection with original index
    o, oi = interval_list.intersect_with_index(3, 15, 'a')
    assert o.size == 2 and o.label_ranges['a'][0] == 10 and o.label_ranges['a'][1] == 13
    assert (oi == np.array([0,3], dtype=int)).all()


def test_wps(interval_list):
    # Expected results
    expected_first_wps = np.array([-2., -2., -2., -2., -2., -1., -1., -2., -3., -3.])
    expected_last_wps = np.array([ 2.,  2.,  2.,  0.,  0.,  0.,  0.,  1., -1., -1.])

    # Test WPS calculation
    w = interval_list.wps(4)
    assert (w[:10] == expected_first_wps).all() and (w[-10:] == expected_last_wps).all()


def test_merge(interval_list):
    # Test merging
    m = interval_list.merge()
    assert m.size == 8 and m.label_ranges['a'][0] == 10 and m.label_ranges['a'][1] == 100


def test_ops(interval_list):
    # Expected results
    expected_subtract_starts = np.array([0, 13, 25, 100, 0, 25, 110], dtype=np.intc)
    expected_subtract_ends = np.array([10, 15, 30, 105, 5, 27, 120], dtype=np.intc)
    expected_common_starts = np.array([10, 11, 15, 19, 30, 20], dtype=np.intc)
    expected_common_ends = np.array([11, 13, 19, 25, 100, 25], dtype=np.intc)

    # Test subtract
    j = IntervalArray()
    j.add(20,27)
    j.add(0,105)
    j.add(0,5)
    j.add(110,120)
    s = j - interval_list
    for k, x in enumerate(s):
        assert x.start == expected_subtract_starts[k]
        assert x.end == expected_subtract_ends[k]

    # Test common
    c = j + interval_list
    for k, x in enumerate(c):
        assert x.start == expected_common_starts[k]
        assert x.end == expected_common_ends[k]


def test_filter(interval_list):
    # Test filtering
    f = interval_list.filter(3, 10)
    assert f.size == 2 and f.first == 15 and f.last == 25


def test_pickle(interval_list):
    # Test pickling
    d = pickle.dumps(interval_list)
    i2 = pickle.loads(d)
    assert i2.size == 7 and i2.first == 10 and i2.last == 100

    # Test pickle adding
    i2.add(40, 60)
    assert len(i2) == 8

    # Test pickle iteration
    is_Interval = 0
    for x in i2:
        is_Interval += isinstance(x, Interval)
    assert is_Interval == 8


def test_length_dist(interval_list):
    # Expected results
    expected_ld = np.array([1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0,
                    0, 0, 0, 1])

    # Test length distribution
    ld = interval_list.length_dist()
    assert (ld == expected_ld).all()


def test_index(interval_list):
    # Test indexing
    assert interval_list[-2].start == 30
    



# ailist version: 1.0.0
from ailist import IntervalArray
# ncls version: 0.0.53
from ncls import NCLS
# numpy version: 1.19.1
import numpy as np
# pandas version: 1.1.0
import pandas as pd
# quicksect version: 0.2.2
import quicksect

# Set seed
np.random.seed(100)


# First values
starts1 = np.random.randint(0, 100000, 100000)
ends1 = starts1 + np.random.randint(1, 10000, 100000)
ids1 = np.arange(len(starts1))

# Second values
starts2 = np.random.randint(0, 100000, 100000)
ends2 = starts2 + np.random.randint(1, 10000, 100000)
ids2 = np.arange(len(starts2))

# Test AIList
i = IntervalArray()

i.from_array(starts1, ends1)
i.construct()

ai_res = i.intersect_from_array(starts2, ends2)

i.intersect(starts2[50], ends2[50])


# Test NCLS
n = NCLS(starts1, ends1, ids1)

n_res = n.all_overlaps_both(starts2, ends2, ids2)

list(n.find_overlap(starts2[50], ends2[50]))

# Test pandas
p = pd.IntervalIndex.from_tuples(list(zip(starts1, ends1)))

p.overlaps(pd.Interval(starts2[50], ends2[50]))

# Test quicksect
b = quicksect.IntervalTree()
for i in range(len(starts1)):
    b.add(starts1[i], ends1[i])

b.search(starts2[50], ends2[50])



#cython: embedsignature=True
#cython: profile=False

import os
import sys
import numpy as np
cimport numpy as np
import math
cimport cython
import pandas as pd
from libc.string cimport memcpy
np.import_array()
from time import time

# Set byteorder for __reduce__
byteorder = sys.byteorder


def get_include():
	"""
	Get file directory if C headers

	Parameters
	----------
		None

	Returns
	-------
		location : str
			Directory to header files
	"""

	# Grab file location
	location = os.path.split(os.path.realpath(__file__))[0]

	return location


cpdef IntervalArray rebuild_IntervalArray(bytes data, bytes b_length):
	"""
	Rebuild function for __reduce__()

	Parameters
	----------
		data : bytes 
			Bytes representation of aiarray_t
		b_length : bytes 
			Length of aiarray_t

	Returns
	-------
		aia : aiarray_t* 
			Translated aiarray_t from data

	"""

	# Initialize new IntervalArray
	aia = IntervalArray()

	# Build IntervalArray from serialized data
	cdef aiarray_t *interval_list = aia._set_data(data, b_length)
	aia.set_list(interval_list)

	return aia


@cython.auto_pickle(True)
cdef class IntervalArray(object):
	"""
	Wrapper for C aiarray_t

	:class:`~IntervalArray.IntervalArray` stores a list of intervals
	"""

	def __cinit__(self):
		"""
		Initialize IntervalArray object
		"""

		self.c_aiarray = aiarray_init()
		self.is_constructed = False
		self.is_closed = False
		self.is_frozen = False


	def __init__(self):
		"""
		Initialize IntervalArray object

		Parameters
		----------
			None

		Returns
		-------
			None

		"""

		pass


	def __dealloc__(self):
		"""
		Free IntervalArray.c_aiarray
		"""
		
		#if hasattr(self, "interval_list"):
			#aiarray_destroy(self.c_aiarray)

		aiarray_destroy(self.c_aiarray)


	cdef bytes _get_data(self):
		"""
		Function to convert aiarray_t to bytes
		for serialization by __reduce__()
		"""

		return <bytes>(<char*>self.c_aiarray.interval_list)[:(sizeof(interval_t)*self.c_aiarray.nr)]


	cdef aiarray_t *_set_data(self, bytes data, bytes b_length):
		"""
		Function to build aiarray_t object from
		serialized bytes using __reduce__()

		Parameters
		----------
			data : bytes 
				Bytes representation of aiarray_t
			b_length : bytes
				Length of aiarray_t

		Returns
		---------
			interval_list : aiarray_t*
				Translated aiarray_t for bytes
		"""
		
		# Convert bytes to ints
		cdef int length = int.from_bytes(b_length, byteorder)
		
		# Create new aiarray_t
		cdef aiarray_t *aia = aiarray_init()
		cdef interval_t *interval_list = <interval_t*>malloc(length * sizeof(interval_t))
		memcpy(interval_list, <char*>data, sizeof(interval_t) * length)

		# Iteratively add intervals to interval_list		
		cdef int i
		for i in range(length):
			aiarray_add(aia, interval_list[i].start, interval_list[i].end)

		return aia


	def __reduce__(self):
		"""
		Used for pickling. Convert IntervalArray to bytes and back.
		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("IntervalArray object has been closed.")
		
		# Convert ints to bytes
		b_length = int(self.c_aiarray.nr).to_bytes(4, byteorder)

		# Convert aiarray_t to bytes
		data = self._get_data()

		return (rebuild_IntervalArray, (data, b_length))

	
	@property
	def nc(self):
		"""
		Number of components in constructed IntervalArray
		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("IntervalArray object has been closed.")

		# Check if object is constructed
		if self.is_constructed:
			return self.c_aiarray.nc
		else:
			return None


	@property
	def lenC(self):
		"""
		Length of components in constructed IntervalArray
		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("IntervalArray object has been closed.")

		# Check if object is constructed
		if self.is_constructed:
			return self.c_aiarray.lenC[self.c_aiarray.nc]
		else:
			return None

	
	@property
	def idxC(self):
		"""
		Index of components in constructed IntervalArray
		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("IntervalArray object has been closed.")

		# Check
		if self.is_constructed:
			return self.c_aiarray.idxC[self.c_aiarray.nc]
		else:
			return None


	@property	
	def size(self):
		"""
		Number of intervals in IntervalArray
		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("IntervalArray object has been closed.")

		return self.c_aiarray.nr
	
	@property
	def first(self):
		"""
		Start of first interval in IntervalArray
		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("IntervalArray object has been closed.")
		
		# Check if ther are any intervals
		if self.size == 0:
			return None
		else:
			return self.c_aiarray.first

	@property
	def last(self):
		"""
		End of last intervals in IntervalArray
		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("IntervalArray object has been closed.")

		# Check if there are any intervals
		if self.size == 0:
			return None
		else:
			return self.c_aiarray.last

	@property
	def range(self):
		"""
		IntervalArray.last - IntervalArray.first
		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("IntervalArray object has been closed.")

		# Check if ther are any intervals
		if self.size == 0:
			return 0
		else:
			return self.last - self.first

	@property
	def id_index(self):
		"""
		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("IntervalArray object has been closed.")
		# Check if object is constructed
		if self.is_constructed == False:
			raise NameError("IntervalArray object must be constructed.")		
		
		ids = np.zeros(len(self), dtype=np.intc)
		cdef int i
		for i in range(len(self)):
			ids[i] = self.c_aiarray.id_index[i]

		return ids

	def __len__(self):
		"""
		Return size of interval_list
		"""
		
		# Check if object is still open
		if self.is_closed:
			raise NameError("IntervalArray object has been closed.")

		return self.size

	
	def __iter__(self):
		"""
		Iterate over IntervalArray object
		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("IntervalArray object has been closed.")

		# Iterate over interval list
		cdef Interval interval
		cdef interval_t *cinterval
		cdef int i
		for i in range(self.size):
			interval = Interval()
			cinterval = aiarray_get_index(self.c_aiarray, i)
			interval.set_i(cinterval)
			
			yield interval


	def __sub__(self, IntervalArray query_ail):
		"""
		Subtract values
		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("IntervalArray object has been closed.")

		return self.subtract(query_ail)


	def __add__(self, IntervalArray query_ail):
		"""
		Common values
		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("IntervalArray object has been closed.")

		return self.common(query_ail)


	def __or__(self, IntervalArray query_ail):
		"""
		Common values
		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("IntervalArray object has been closed.")

		return self.append(query_ail)

	
	def __hash__(self):
		"""
		Get hash value
		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("IntervalArray object has been closed.")

		return hash(self)
		
		
	cdef aiarray_t *_array_index(IntervalArray self, const long[::1] ids):
		"""
		Index and array looking for index value

		Parameters
		----------
			self : IntervalArray
				self object
			ids : const long[::1]
				Indices

		Returns
		-------
			cindexed_IntervalArray : aiarray_t
				Subseted array

		"""
		
		# Define variables
		cdef int length = len(ids)
		cdef aiarray_t *cindexed_IntervalArray
		
		# Call C indexing to subset aiarray_t
		cindexed_IntervalArray = aiarray_array_index(self.c_aiarray, &ids[0], length)
		
		return cindexed_IntervalArray

	def __getitem__(self, key):
		"""
		Index Intervals by value
		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("IntervalArray object has been closed.")

		# Define variables
		cdef IntervalArray indexed_IntervalArray
		cdef aiarray_t *cindexed_IntervalArray
		cdef int k
		cdef interval_t *i
		cdef int slice_start
		cdef int slice_end
		cdef int slice_step

		# Check if key is iterable
		try:
			iter(key) # Test is present
			# Iterate over key
			indexed_IntervalArray = IntervalArray()

			# Check if keys are booleans
			if isinstance(key[0], np.bool_):
				for k in range(len(key)):
					# if True
					if key[k]:
						# Add interval
						#i = self.c_aiarray.interval_list[k]
						i = aiarray_get_index(self.c_aiarray, k)
						indexed_IntervalArray.add(i.start, i.end)

			# Must be integers
			else:
				cindexed_IntervalArray = self._array_index(key)
				indexed_IntervalArray.set_list(cindexed_IntervalArray)
			
			return indexed_IntervalArray
		
		# key is not iterable, treat as int
		except TypeError:
			# Check if key is slice
			if isinstance(key, slice):
				# Determine indices
				slice_start, slice_end, slice_step = key.indices(self.size)
				# Iterate over key
				indexed_IntervalArray = IntervalArray()
				for k in range(slice_start, slice_end, slice_step):
					# Add interval
					#i = self.c_aiarray.interval_list[k]
					i = aiarray_get_index(self.c_aiarray, k)
					indexed_IntervalArray.add(i.start, i.end)
					
				return indexed_IntervalArray

			# Check if key is greater than length
			if key >= self.c_aiarray.nr:
				raise IndexError("Value larger than IntervalArray length")

			# Check if negative
			if key < 0:
				key = self.c_aiarray.nr + key

			# Create Interval wrapper
			output_interval = Interval()
			i = aiarray_get_index(self.c_aiarray, key)
			output_interval.set_i(i)
		
		return output_interval


	def __repr__(self):
		"""
		Representation of IntervalArray object
		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("IntervalArray object has been closed.")

		# Initialize string
		repr_string = "IntervalArray\n"
		if self.size == 0:
			repr_string += " range: (None-None)\n"
		else:
			repr_string += " range: (%d-%d)\n" % (self.first, self.last)

		# Iterate over interval_list
		if self.c_aiarray.nr > 10:
			for i in range(5):
				repr_string += "   (%d-%d)\n" % (self[i].start, self[i].end)
			repr_string += "   ...\n"
			for i in range(-5, 0, 1):
				repr_string += "   (%d-%d)\n" % (self[i].start, self[i].end)
		else:
			for i in range(self.c_aiarray.nr):
				repr_string += "   (%d-%d)\n" % (self[i].start, self[i].end)

		return repr_string		


	cdef void set_list(IntervalArray self, aiarray_t *input_list):
		"""
		Set wrapper of C IntervalArray

		Parameters
		----------
			input_list : aiarray_t*
				aiarray_t to replace existing one

		Returns
		-------
			None
		"""

		# Free old skiplist
		#if self.c_aiarray:
			#aiarray_destroy(self.c_aiarray)
		aiarray_destroy(self.c_aiarray)
		
		# Replace new skiplist
		self.c_aiarray = input_list
		self.is_closed = False


	def freeze(self):
		"""
		Make :class:`~IntervalArray.IntervalArray` immutable

		Parameters
		----------
			None

		Returns
		-------
			None

		See Also
		--------
		IntervalArray.unfreeze: Make mutable
		IntervalArray.sort: Sort intervals inplace
		IntervalArray.construct: Construct IntervalArray, required to call IntervalArray.intersect

		Examples
		--------
		>>> from IntervalArray import IntervalArray
		>>> ail = IntervalArray()
		>>> ail.add(1, 2)
		>>> ail.add(3, 4)
		>>> ail.add(3, 6)
		>>> ail
		IntervalArray
		  range: (1-6)
		   (1-2)
		   (3-4)
		   (3-6)
		>>> ail.freeze()
		>>> ail.add(9, 10)
		TypeError: IntervalArray is frozen and currently immutatable. Try '.unfreeze()' to reverse.

		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("IntervalArray object has been closed.")

		# Make sure it is constructed
		if self.is_constructed == False:
			self.construct()

		# Change to frozen
		self.is_frozen = True


	def unfreeze(self):
		"""
		Make :class:`~IntervalArray.IntervalArray` mutable

		Parameters
		----------
			None

		Returns
		-------
			None

		See Also
		--------
		IntervalArray.freeze: Make immutable
		IntervalArray.sort: Sort intervals inplace
		IntervalArray.construct: Construct IntervalArray, required to call IntervalArray.intersect

		Examples
		--------
		>>> from IntervalArray import IntervalArray
		>>> ail = IntervalArray()
		>>> ail.add(1, 2)
		>>> ail.add(3, 4)
		>>> ail.add(3, 6)
		>>> ail
		IntervalArray
		  range: (1-6)
		   (1-2)
		   (3-4)
		   (3-6)
		>>> ail.freeze()
		>>> ail.add(9, 10)
		TypeError: IntervalArray is frozen and currently immutatable. Try '.unfreeze()' to reverse.
		>>> ail.unfreeze()
		>>> ail.add(9, 10)

		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("IntervalArray object has been closed.")

		# Change to not frozen
		self.is_frozen = False


	cdef void _insert(IntervalArray self, int start, int end):
		aiarray_add(self.c_aiarray, start, end)

	def add(self, int start, int end):
		"""
		Add an interval to IntervalArray inplace
		
		Parameters
		----------
			start : int
				Start position of interval
			end : int
				End position of interval

		Returns
		-------
			None

		See Also
		--------
		IntervalArray.from_array: Add intervals from arrays
		IntervalArray.construct: Construct IntervalArray, required to call IntervalArray.intersect
		IntervalArray.intersect: Find intervals overlapping given range

		Examples
		--------
		>>> from IntervalArray import IntervalArray
		>>> ail = IntervalArray()
		>>> ail.add(1, 2)
		>>> ail.add(3, 4)
		>>> ail.add(3, 6)
		>>> ail
		IntervalArray
		  range: (1-6)
		   (1-2)
		   (3-4)
		   (3-6)

		"""
		
		# Check if object is still open
		if self.is_closed:
			raise NameError("IntervalArray object has been closed.")

		# Check that object is not frozen
		if self.is_frozen:
			raise TypeError("IntervalArray is frozen and currently immutatable. Try '.unfreeze()' to reverse.")

		# Insert interval
		self._insert(start, end)
		self.is_constructed = False


	cdef np.ndarray _get_comp_bounds(IntervalArray self):
		"""
		Get component index for label
		"""

		# Initialize label specific variables
		cdef int *idxC = self.c_aiarray.idxC
		cdef int n_comps = self.c_aiarray.nc
		cdef np.ndarray comps_bounds = np.zeros(n_comps + 1, dtype=int)

		# Iterate over components
		cdef int i
		for i in range(n_comps):
			comps_bounds[i] = idxC[i]
		comps_bounds[n_comps] = self.c_aiarray.nr

		return comps_bounds
	
	def iter_sorted(self):
		"""
		Iterate over an IntervalArray in sort by starts

		Parameters
		----------
			None

		Returns
		-------
			intervals : generator of Interval objects
				Generates Interval objects sorted by starts

		"""

		#raise NotImplementedError("This feature is not implemented yet")

		# Check if object is still open
		if self.is_closed:
			raise NameError("IntervalArray object has been closed.")

		# Check if object is constructed
		if not self.is_constructed:
			raise TypeError("IntervalArray object must be constructed first. Try '.construct()' first.")

		# Define variables
		#cdef np.ndarray self.ail.idxC[]
		cdef np.ndarray comp_bounds
		cdef np.ndarray label_comp_used
		cdef int n
		cdef int position
		cdef int selected_comp
		#cdef int label_start
		#cdef int label_end
		cdef interval_t *cintv
		cdef Interval output_interval

		#label_code = self.label_map[label]
		comp_bounds = self._get_comp_bounds()
		comp_used = comp_bounds[:len(comp_bounds) - 1].copy()

		#label_start = self.ail.label_index[label_code]
		#label_end = self.ail.label_index[label_code + 1]

		# Iterate over component intervals
		cintv = &self.c_aiarray.interval_list[0]
		for n in range(self.size):
			selected_comp = 0
			# Iterate over other components
			for j in range(len(comp_bounds) - 1):
				# Check component has intervals left to investigate
				if comp_used[j] == comp_bounds[j + 1]:
					continue
				# Determine position
				position = comp_used[j]
				# Check for lower start
				if self.c_aiarray.interval_list[position].start < cintv.start:
					cintv = &self.c_aiarray.interval_list[position]
					selected_comp = j

			# Create Interval wrapper
			output_interval = Interval()
			output_interval.set_i(cintv)
			yield output_interval
			
			# Increment comp_counter for selected comp
			comp_used[selected_comp] += 1
			# Iterate over components
			for j in range(len(comp_bounds) - 1):
				# If position is available, assign next interval
				if comp_used[j] != comp_bounds[j + 1]:
					position = comp_used[j]
					cintv = &self.c_aiarray.interval_list[position]
					break


	def from_array(self, const long[::1] starts, const long[::1] ends):
		"""
		Add intervals from arrays to IntervalArray inplace
		
		Parameters
		----------
			starts : ~numpy.ndarray{long}
				Start positions of intervals
			ends : numpy.ndarray{long}
				End positions of intervals

		Returns
		-------
			None

		See Also
		--------
		IntervalArray.add: Add interval to IntervalArray
		IntervalArray.construct: Construct IntervalArray, required to call IntervalArray.intersect
		IntervalArray.intersect: Find intervals overlapping given range

		Examples
		--------
		>>> from IntervalArray import IntervalArray
		>>> import numpy as np
		>>> starts = np.arange(100)
		>>> ends = starts + 10
		>>> ail = IntervalArray()
		>>> ail.from_array(starts, ends)
		>>> ail.
		IntervalArray
		  range: (0-109)
		   (0-10)
		   (1-11)
		   (2-12)
		   (3-13)
		   (4-14)
		   ...
		   (95-105)
		   (96-106)
		   (97-107)
		   (98-108)

		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("IntervalArray object has been closed.")

		# Check that object is not frozen
		if self.is_frozen:
			raise TypeError("IntervalArray is frozen and currently immutatable. Try '.unfreeze()' to reverse.")

		# Call C function to create
		cdef int array_length = len(starts)
		aiarray_from_array(self.c_aiarray, &starts[0], &ends[0], array_length)
		self.is_constructed = False


	cdef void _construct(IntervalArray self, int min_length):
		# Contruct
		aiarray_construct(self.c_aiarray, min_length)
		# Remember input order
		aiarray_cache_id(self.c_aiarray)

	def construct(self, int min_length=20):
		"""
		Construct aiarray_t *Required to call intersect

		Parameters
		----------
			min_length : int
				Minimum length

		Returns
		-------
			None

		.. warning::
			This will re-sort intervals inplace

		See Also
		--------
		IntervalArray.sort: Sort intervals inplace
		IntervalArray.intersect: Find intervals overlapping given range

		Examples
		--------
		>>> from IntervalArray import IntervalArray
		>>> ail = IntervalArray()
		>>> ail.add(1, 2)
		>>> ail.add(3, 4)
		>>> ail.add(2, 6)
		>>> ail
		IntervalArray
		  range: (1-6)
		   (1-2)
		   (3-4)
		   (2-6)
		>>> ail.construct()
		>>> ail
		IntervalArray
		  range: (1-6)
		   (1-2)
		   (2-6)
		   (3-4)

		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("IntervalArray object has been closed.")

		# Check if already constructed
		if self.is_constructed == False:
			self._construct(min_length)
			self.is_constructed = True
		else:
			pass


	cdef aiarray_t *_intersect(IntervalArray self, int start, int end):
		cdef aiarray_t *overlaps = aiarray_query(self.c_aiarray, start, end)

		return overlaps

	def intersect(self, int start, int end):
		"""
		Find intervals overlapping given range
		
		Parameters
		----------
			start : int
				Start position of query range
			end : int
				End position of query range

		Returns
		-------
			overlaps : IntervalArray
				Overlapping intervals

		.. warning::
			This requires :func:`~IntervalArray.IntervalArray.construct` and will run it if not already run. This will re-sort intervals inplace.

		See Also
		--------
		IntervalArray.construct: Construct IntervalArray, required to call IntervalArray.intersect
		IntervalArray.add: Add interval to IntervalArray
		IntervalArray.intersect_from_array: Find interval indices overlapping given ranges

		Examples
		--------
		>>> from IntervalArray import IntervalArray
		>>> ail = IntervalArray()
		>>> ail.add(1, 2)
		>>> ail.add(3, 4)
		>>> ail.add(2, 6)
		>>> ail
		IntervalArray
		  range: (1-6)
		   (1-2)
		   (3-4)
		   (2-6)
		>>> q = ail.intersect(2, 10)
		>>> q
		IntervalArray
		  range: (2-6)
		   (2-6)
		   (3-4)

		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("IntervalArray object has been closed.")

		# Check if is constructed
		if self.is_constructed == False:
			self.construct()

		# Intersect
		cdef aiarray_t *i_list = self._intersect(start, end)
		cdef IntervalArray overlaps = IntervalArray()
		overlaps.set_list(i_list)

		return overlaps


	@cython.boundscheck(False)
	@cython.wraparound(False)
	@cython.initializedcheck(False)
	cpdef _intersect_from_array(IntervalArray self, const long[::1] starts, const long[::1] ends):
		cdef int length = len(starts)
		cdef array_query_t *total_overlaps
		total_overlaps = aiarray_query_from_array(self.c_aiarray, &starts[0], &ends[0], length)

		cdef np.ndarray ref_index = pointer_to_numpy_array(total_overlaps.ref_index, total_overlaps.size)
		cdef np.ndarray query_index = pointer_to_numpy_array(total_overlaps.query_index, total_overlaps.size)

		return ref_index, query_index

	def intersect_from_array(self, const long[::1] starts, const long[::1] ends):
		"""
		Find interval indices overlapping given ranges
		
		Parameters
		----------
			starts : numpy.ndarray{long}
				Start positions of intervals
			ends : numpy.ndarray{long}
				End positions of intervals

		Returns
		-------
			ref_index : np.ndarray{int}
				Overlapping interval indices from IntervalArray
			query_index : np.ndarray{int}
				Overlapping interval indices from query IntervalArray

		.. warning::
			This requires :func:`~IntervalArray.IntervalArray.construct` and will run it if not already run. This will re-sort intervals inplace.

		See Also
		--------
		IntervalArray.construct: Construct IntervalArray, required to call IntervalArray.intersect
		IntervalArray.add: Add interval to IntervalArray
		IntervalArray.intersect: Find intervals overlapping given range

		Examples
		--------
		>>> from IntervalArray import IntervalArray
		>>> ail1 = IntervalArray()
		>>> ail1.add(1, 2)
		>>> ail1.add(3, 4)
		>>> ail1.add(2, 6)
		>>> ail1
		IntervalArray
		  range: (1-6)
		   (1-2)
		   (3-4)
		   (2-6)
		>>> ail2 = IntervalArray()
		>>> ail2.add(1, 2)
		>>> ail2.add(3, 6)
		>>> ail2
		IntervalArray
		  range: (1-6)
		    (1-2)
		    (3-6)
		>>> q = ail1.intersect_from_array(ail2)
		>>> q
		(array([2, 1]), array([]))

		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("IntervalArray object has been closed.")

		# Check if object is constructed
		if self.is_constructed == False:
			self.construct()

		# Call C function
		ref_index, query_index = self._intersect_from_array(starts, ends)
		
		return ref_index, query_index


	cpdef _intersect_from_aiarray(IntervalArray self, IntervalArray ail):
		# Intersect with other IntervalArray
		cdef array_query_t *total_overlaps = aiarray_query_from_aiarray(self.c_aiarray, ail.c_aiarray)

		# Create numpy array from C pointer
		cdef np.ndarray ref_index = pointer_to_numpy_array(total_overlaps.ref_index, total_overlaps.size)
		cdef np.ndarray query_index = pointer_to_numpy_array(total_overlaps.query_index, total_overlaps.size)

		return ref_index, query_index

	def intersect_from_IntervalArray(self, IntervalArray ail_query):
		"""
		Find interval indices overlapping given ranges
		
		Parameters
		----------
			ail_query : IntervalArray
				Intervals to query

		Returns
		-------
			ref_index : np.ndarray{int}
				Overlapping interval indices from IntervalArray
			query_index : np.ndarray{int}
				Overlapping interval indices from query IntervalArray

		See Also
		--------
		IntervalArray.construct: Construct IntervalArray, required to call IntervalArray.intersect
		IntervalArray.add: Add interval to IntervalArray
		IntervalArray.intersect: Find intervals overlapping given range
		IntervalArray.intersect_from_array: Find interval indices overlapping given range

		Examples
		--------
		>>> from IntervalArray import IntervalArray
		>>> ail1 = IntervalArray()
		>>> ail1.add(1, 2)
		>>> ail1.add(3, 4)
		>>> ail1.add(2, 6)
		>>> ail1
		IntervalArray
		  range: (1-6)
		   (1-2, 0)
		   (3-4, 1)
		   (2-6, 2)
		>>> ail2 = IntervalArray()
		>>> ail2.add(1, 2)
		>>> ail2.add(3, 6)
		>>> ail2
		IntervalArray
		  range: (1-6)
		    (1-2, 0)
		    (3-6, 1)
		>>> q = ail1.intersect_from_IntervalArray(ail2)
		>>> q
		(array([0, 1, 1]), array([0, 2, 1]))

		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("IntervalArray object has been closed.")

		# Check if object is constructed
		if self.is_constructed == False:
			self.construct()

		# Intersect
		ref_index, query_index = self._intersect_from_aiarray(ail_query)
		
		return ref_index, query_index


	cpdef _intersect_with_index(IntervalArray self, int start, int end):
		# Intersect with other IntervalArray
		cdef overlap_index_t *total_overlaps = aiarray_query_with_index(self.c_aiarray, start, end)

		# Create numpy array from C pointer
		cdef np.ndarray indices = pointer_to_numpy_array(total_overlaps.indices, total_overlaps.size)
		cdef IntervalArray overlaps = IntervalArray()
		overlaps.set_list(total_overlaps.ail)

		return overlaps, indices
	
	def intersect_with_index(self, int start, int end):
		"""
		Find interval indices overlapping given ranges
		
		Parameters
		----------
			ail_query : IntervalArray
				Intervals to query

		Returns
		-------
			ref_index : np.ndarray{int}
				Overlapping interval indices from IntervalArray
			query_index : np.ndarray{int}
				Overlapping interval indices from query IntervalArray

		See Also
		--------
		IntervalArray.construct: Construct IntervalArray, required to call IntervalArray.intersect
		IntervalArray.add: Add interval to IntervalArray
		IntervalArray.intersect: Find intervals overlapping given range
		IntervalArray.intersect_from_array: Find interval indices overlapping given range

		Examples
		--------
		>>> from IntervalArray import IntervalArray
		>>> ail1 = IntervalArray()
		>>> ail1.add(1, 2)
		>>> ail1.add(3, 4)
		>>> ail1.add(2, 6)
		>>> ail1
		IntervalArray
		  range: (1-6)
		   (1-2)
		   (3-4)
		   (2-6)
		>>> ail2 = IntervalArray()
		>>> ail2.add(1, 2)
		>>> ail2.add(3, 6)
		>>> ail2
		IntervalArray
		  range: (1-6)
		    (1-2)
		    (3-6)
		>>> q = ail1.intersect_from_IntervalArray(ail2)
		>>> q
		(array([0, 1, 1]), array([0, 2, 1]))

		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("IntervalArray object has been closed.")

		# Check if object is constructed
		if self.is_constructed == False:
			self.construct()

		# Intersect
		cdef IntervalArray overlaps
		cdef np.ndarray indices
		overlaps, indices = self._intersect_with_index(start, end)
		
		return overlaps, indices
		

	cdef np.ndarray _coverage(IntervalArray self):
		# Initialize coverage
		cdef double[::1] coverage = np.zeros(self.range, dtype=np.double)

		# Call C function
		aiarray_coverage(self.c_aiarray, &coverage[0])

		return np.asarray(coverage)

	def coverage(self):
		"""
		Find number of intervals overlapping every
		position in the AList range

		Parameters
		----------
			None

		Returns
		-------
			coverage : pandas.Series{double}
				Position on index and coverage as values

		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("IntervalArray object has been closed.")

		# Initialize coverage
		cdef np.ndarray coverage
		# Calculate coverage
		coverage = self._coverage()
		
		return pd.Series(coverage, index=np.arange(self.first, self.last))


	cdef np.ndarray _bin_coverage(IntervalArray self, int bin_size):
		# Initialize coverage
		cdef int n_bins = math.ceil(self.last / bin_size) - (self.first // bin_size)
		cdef double[::1] bins = np.zeros(n_bins, dtype=np.double)

		# Call C function
		aiarray_bin_coverage(self.c_aiarray, &bins[0], bin_size)

		return np.asarray(bins)

	cdef np.ndarray _bin_coverage_length(IntervalArray self, int bin_size, int min_length, int max_length):
		# Initialize coverage
		cdef int n_bins = math.ceil(self.last / bin_size) - (self.first // bin_size)
		cdef double[::1] bins = np.zeros(n_bins, dtype=np.double)

		# Call C function
		aiarray_bin_coverage_length(self.c_aiarray, &bins[0], bin_size, min_length, max_length)

		return np.asarray(bins)

	def bin_coverage(self, int bin_size=100000, min_length=None, max_length=None):
		"""
		Find sum of coverage within binned
		positions
		
		Parameters
		----------
			bin_size : int
				Size of the bin to use
			min_length : int
				Minimum length of intervals to include [default = None]
			max_length : int
				Maximum length of intervals to include [default = None]

		Returns
		-------
			bins : pandas.Series{double}
				Position on index and coverage as values

		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("IntervalArray object has been closed.")

		# Initialize coverage
		cdef np.ndarray bins
		# Calculate coverage
		if min_length is None or max_length is None:
			bins = self._bin_coverage(bin_size)
		else:
			bins = self._bin_coverage_length(bin_size, min_length, max_length)
		
		return pd.Series(bins, index=(np.arange(len(bins)) + int(self.first / bin_size)) * bin_size)


	cdef np.ndarray _bin_nhits(IntervalArray self, int bin_size):
		# Initialize coverage
		cdef int n_bins = math.ceil(self.last / bin_size) - (self.first // bin_size)
		cdef double[::1] bins = np.zeros(n_bins, dtype=np.double)
		
		# Call C function
		aiarray_bin_nhits(self.c_aiarray, &bins[0], bin_size)

		return np.asarray(bins)

	cdef np.ndarray _bin_nhits_length(IntervalArray self, int bin_size, int min_length, int max_length):
		# Initialize coverage
		cdef int n_bins = math.ceil(self.last / bin_size) - (self.first // bin_size)
		cdef double[::1] bins = np.zeros(n_bins, dtype=np.double)

		# Call C function
		aiarray_bin_nhits_length(self.c_aiarray, &bins[0], bin_size, min_length, max_length)

		return np.asarray(bins)

	def bin_nhits(self, int bin_size=100000, min_length=None, max_length=None):
		"""
		Find number of intervals overlapping binned
		positions
		
		Parameters
		----------
			bin_size : int
				Size of the bin to use
			min_length : int
				Minimum length of intervals to include [default = None]
			max_length : int
				Maximum length of intervals to include [default = None]

		Returns
		-------
			bins : pandas.Series{double}
				Position on index and coverage as values

		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("IntervalArray object has been closed.")

		# Initialize coverage
		cdef np.ndarray bins
		# Calculate coverage
		if min_length is None or max_length is None:
			bins = self._bin_nhits(bin_size)
		else:
			bins = self._bin_nhits_length(bin_size, min_length, max_length)
		
		return pd.Series(bins, index=(np.arange(len(bins)) + int(self.first / bin_size)) * bin_size)


	def display(self):
		"""
		Print all intervals
		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("IntervalArray object has been closed.")

		# Call C function
		display_array(self.c_aiarray)


	def merge(self, int gap=0):
		"""
		Merge intervals within a gap
		
		Parameters
		----------
			gap : int
				Gap between intervals to merge

		Returns
		-------
			merged_list : IntervalArray
				Merged intervals

		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("IntervalArray object has been closed.")

		# Make sure list is constructed
		if self.is_constructed == False:
			self.construct()

		# Create merged
		cdef IntervalArray merged_list = IntervalArray()
		# Call C function
		cdef aiarray_t *merged_clist = aiarray_merge(self.c_aiarray, gap)
		merged_list.set_list(merged_clist)

		return merged_list


	def subtract(self, IntervalArray query_ail):
		"""
		Subtract intervals within another IntervalArray
		
		Parameters
		----------
			query_ail : IntervalArray
				IntervalArray of intervals to subtract

		Returns
		-------
			subtracted_list : IntervalArray
				Subtracted intervals
		"""

		# Check if object is still open
		if self.is_closed or query_ail.is_closed:
			raise NameError("IntervalArray object has been closed.")

		# Make sure list is constructed
		if self.is_constructed == False:
			self.construct()
		if query_ail.is_constructed == False:
			query_ail.construct()

		# Create subracted
		cdef IntervalArray subtracted_list = IntervalArray()
		cdef aiarray_t *subtracted_clist = aiarray_subtract(query_ail.c_aiarray,
														  self.c_aiarray)
		subtracted_list.set_list(subtracted_clist)

		return subtracted_list


	def common(self, IntervalArray query_ail):
		"""
		Common intervals within another IntervalArray
		
		Parameters
		----------
			query_ail : IntervalArray
				IntervalArray of intervals to find commons

		Returns
		-------
			common_list : IntervalArray
				Common intervals

		"""

		# Check if object is still open
		if self.is_closed or query_ail.is_closed:
			raise NameError("IntervalArray object has been closed.")

		# Make sure list is constructed
		if self.is_constructed == False:
			self.construct()
		if query_ail.is_constructed == False:
			query_ail.construct()

		# Create common
		cdef IntervalArray common_list = IntervalArray()
		cdef aiarray_t *common_clist = aiarray_common(query_ail.c_aiarray,
													self.c_aiarray)
		common_list.set_list(common_clist)

		return common_list


	def append(self, IntervalArray query_ail):
		"""
		Union of intervals within two IntervalArray
		
		Parameters
		----------
			query_ail : IntervalArray
				IntervalArray of intervals to append

		Returns
		-------
			union_list: IntervalArray
				Union of intervals

		"""

		# Check if object is still open
		if self.is_closed or query_ail.is_closed:
			raise NameError("IntervalArray object has been closed.")

		# Create union
		cdef IntervalArray union_list = IntervalArray()
		cdef aiarray_t *union_clist = aiarray_append(query_ail.c_aiarray,
												   self.c_aiarray)
		union_list.set_list(union_clist)

		return union_list


	cdef np.ndarray _wps(IntervalArray self, int protection):
		# Initialize wps
		cdef double[::1] wps = np.zeros(self.range, dtype=np.double)

		aiarray_wps(self.c_aiarray, &wps[0], protection)

		return np.asarray(wps)

	cdef np.ndarray _wps_length(IntervalArray self, int protection, int min_length, int max_length):
		# Initialize wps
		cdef double[::1] wps = np.zeros(self.range, dtype=np.double)

		aiarray_wps_length(self.c_aiarray, &wps[0], protection, min_length, max_length)

		return np.asarray(wps)

	def wps(self, int protection=60, min_length=None, max_length=None):
		"""
		Calculate Window Protection Score
		for each position in IntervalArray range
		
		Parameters
		----------
			protection : int
				Protection window to use
			min_length : int
				Minimum length of intervals to include [default = None]
			max_length : int
				Maximum length of intervals to include [default = None]

		Returns
		-------
			wps : pandas.Series{double}
				Position on index and WPS as values

		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("IntervalArray object has been closed.")
		
		# Initialize wps
		cdef np.ndarray wps
		# Calculate wps
		if self.range == 0:
			return None
		if min_length is None or max_length is None:
			wps = self._wps(protection)
		else:
			wps = self._wps_length(protection, min_length, max_length)
		
		return pd.Series(wps, index=np.arange(self.first, self.last))

	
	def filter(self, int min_length=1, int max_length=400):
		"""
		Filter out intervals outside of a length range
		
		Parameters
		----------
			min_length : int
				Minimum length to keep
			max_length : int
				Maximum langth to keep

		Returns
		-------
			filtered_ail : IntervalArray
				Filtered intervals

		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("IntervalArray object has been closed.")

		# Initialize filtered list
		cdef IntervalArray filtered_ail = IntervalArray()

		cdef aiarray_t *cfiltered_ail = aiarray_length_filter(self.c_aiarray, min_length, max_length)
		filtered_ail.set_list(cfiltered_ail)

		return filtered_ail


	cdef np.ndarray _length_dist(IntervalArray self):
		# Initialize distribution
		cdef int max_length = aiarray_max_length(self.c_aiarray)
		cdef int[::1] distribution = np.zeros(max_length + 1, dtype=np.intc)

		# Calculate distribution
		aiarray_length_distribution(self.c_aiarray, &distribution[0])

		return np.asarray(distribution, dtype=np.intc)

	def length_dist(self):
		"""
		Calculate length distribution of intervals

		Parameters
		----------
			None
		
		Returns
		-------
			distribution : numpy.ndarray{int}
				Interval length distribution

		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("IntervalArray object has been closed.")

		# Initialize distribution
		cdef np.ndarray distribution
		# Calculate distribution
		distribution = self._length_dist()

		return distribution


	cdef np.ndarray _nhits_from_array(IntervalArray self, const long[::1] starts, const long[::1] ends):
		# Initialize hits
		cdef int length = starts.size
		cdef int[::1] nhits = np.zeros(length, dtype=np.intc)

		# Calculate distribution
		aiarray_nhits_from_array(self.c_aiarray, &starts[0], &ends[0], length, &nhits[0])

		return np.asarray(nhits, dtype=np.intc)

	cdef np.ndarray _nhits_from_array_length(IntervalArray self, const long[::1] starts, const long[::1] ends, int min_length, int max_length):
		# Initialize hits
		cdef int length = starts.size
		cdef int[::1] nhits = np.zeros(length, dtype=np.intc)

		# Calculate distribution
		aiarray_nhits_from_array_length(self.c_aiarray, &starts[0], &ends[0], length, &nhits[0], min_length, max_length)

		return np.asarray(nhits, dtype=np.intc)

	def nhits_from_array(self, const long[::1] starts, const long[::1] ends, min_length=None, max_length=None):
		"""
		Find number of intervals overlapping given
		positions
		
		Parameters
		----------
			starts : numpy.ndarray{long}
				Start positions to intersect
			ends : numpy.ndarray{long}
				End positions to intersect
			min_length : int
				Minimum length of intervals to include [default = None]
			max_length : int
				Maximum length of intervals to include [default = None]

		Returns
		-------
			nhits : numpy.ndarray{int}
				Number of hits per position

		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("IntervalArray object has been closed.")

		# Make sure list is constructed
		if self.is_constructed == False:
			self.construct()

		# Initialize distribution
		cdef np.ndarray nhits
		# Calculate distribution
		if min_length is None or max_length is None:
			nhits = self._nhits_from_array(starts, ends)
		else:
			nhits = self._nhits_from_array_length(starts, ends, min_length, max_length)

		return nhits


	cdef np.ndarray _interval_coverage(IntervalArray self, int start, int end):
		# Initialize hits
		cdef int[::1] coverage = np.zeros(end - start, dtype=np.intc)

		# Calculate distribution
		aiarray_interval_coverage(self.c_aiarray, start, end, &coverage[0])

		return np.asarray(coverage, dtype=np.intc)
	
	def interval_coverage(self, int start, int end):
		"""
		Find number of intervals overlapping each
		position in given interval
		
		Parameters
		----------
			start : int
				Start position to intersect
			end : int
				End position to intersect

		Returns
		-------
			coverage : pandas.Series{int}
				Number of hits per position

		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("IntervalArray object has been closed.")

		# Make sure list is constructed
		if self.is_constructed == False:
			self.construct()

		# Initialize distribution
		cdef np.ndarray coverage
		# Calculate distribution
		coverage = self._interval_coverage(start, end)

		return pd.Series(coverage, index=np.arange(start, end))

	
	def downsample(self, double proportion):
		"""
		Randomly downsample IntervalArray
		
		Parameters
		----------
			proportion : double
				Proportion of intervals to keep

		Returns
		-------
			filtered_ail : IntervalArray
				Downsampled IntervalArray

		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("IntervalArray object has been closed.")

		# Initialize filtered list
		cdef IntervalArray filtered_ail = IntervalArray()

		cdef aiarray_t *cfiltered_ail = aiarray_downsample(self.c_aiarray, proportion)
		filtered_ail.set_list(cfiltered_ail)

		return filtered_ail

	
	cpdef _downsample_with_index(IntervalArray self, double proportion):
		# Randomly downsample IntervalArray
		cdef overlap_index_t *new_intervals = aiarray_downsample_with_index(self.c_aiarray, proportion)

		# Create numpy array from C pointer
		cdef np.ndarray indices = pointer_to_numpy_array(new_intervals.indices, new_intervals.size)
		cdef IntervalArray intervals = IntervalArray()
		intervals.set_list(new_intervals.ail)

		return intervals, indices
	
	def downsample_with_index(self, double proportion):
		"""
		Find interval indices overlapping given ranges
		
		Parameters
		----------
			ail_query : IntervalArray
				Intervals to query

		Returns
		-------
			ref_index : np.ndarray{int}
				Overlapping interval indices from IntervalArray
			query_index : np.ndarray{int}
				Overlapping interval indices from query IntervalArray

		See Also
		--------
		IntervalArray.construct: Construct IntervalArray, required to call IntervalArray.intersect
		IntervalArray.add: Add interval to IntervalArray
		IntervalArray.intersect: Find intervals overlapping given range
		IntervalArray.intersect_from_array: Find interval indices overlapping given range

		Examples
		--------
		>>> from IntervalArray import IntervalArray
		>>> ail1 = IntervalArray()
		>>> ail1.add(1, 2)
		>>> ail1.add(3, 4)
		>>> ail1.add(2, 6)
		>>> ail1
		IntervalArray
		  range: (1-6)
		   (1-2)
		   (3-4)
		   (2-6)
		>>> ail2 = IntervalArray()
		>>> ail2.add(1, 2)
		>>> ail2.add(3, 6)
		>>> ail2
		IntervalArray
		  range: (1-6)
		    (1-2)
		    (3-6)
		>>> q = ail1.intersect_from_IntervalArray(ail2)
		>>> q
		(array([0, 1, 1]), array([0, 2, 1]))

		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("IntervalArray object has been closed.")

		# Check if object is constructed
		if self.is_constructed == False:
			self.construct()

		# Intersect
		cdef IntervalArray intervals
		cdef np.ndarray indices
		intervals, indices = self._downsamples_with_index(proportion)
		
		return intervals, indices


	def index_by_IntervalArray(self, IntervalArray ail, inplace=False):
		"""
		Use IntervalArray as an index to adjust values given other IntervalArray

		Parameters
		----------
			ail : IntervalArray
				Intervals to index
			inplace : bool
				Whether to due inplace (If done wrong, can lead to corrupted data!)
		
		Returns
		-------
			indexed_ail : IntervalArray
				Re-indexed intervals

		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("IntervalArray object has been closed.")

		# Check if frozen
		if inplace and self.is_frozen:
			raise TypeError("IntervalArray is frozen and currently immutatable. Try '.unfreeze()' to reverse.")

		# Index
		cdef IntervalArray indexed_ail = IntervalArray()
		cdef aiarray_t *cindexed_ail
		cdef exit_code
		if inplace:
			exit_code = aiarray_index_by_aiarray_inplace(self.c_aiarray, ail.c_aiarray)
			# Check if it ran correctly
			if exit_code == 1:
				raise LookupError("Index out of bounds. DATA IS LIKELY CURRUPTED NOW!")
		else:
			cindexed_ail = aiarray_index_by_aiarray(self.c_aiarray, ail.c_aiarray)
			indexed_ail.set_list(cindexed_ail)
			# Check if it ran correctly
			if indexed_ail.size != self.size:
				raise LookupError("Index out of bounds.")

			return indexed_ail


	def extract_starts(self):
		"""
		Return the start values

		Parameters
		----------
			None

		Returns
		-------
			numpy.ndarray
				Start values
		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("IntervalArray object has been closed.")

		# Extract start values
		cdef long[::1] starts = np.zeros(self.size, dtype=np.int_)
		aiarray_extract_starts(self.c_aiarray, &starts[0])

		return np.asarray(starts, dtype=np.intc)

	
	def extract_ends(self):
		"""
		Return the end values

		Parameters
		----------
			None

		Returns
		-------
			numpy.ndarray
				End values
		"""

		# Check if object is still open
		if self.is_closed:
			raise NameError("IntervalArray object has been closed.")

		# Extract end values
		cdef long[::1] ends = np.zeros(self.size, dtype=np.int_)
		aiarray_extract_ends(self.c_aiarray, &ends[0])

		return np.asarray(ends, dtype=np.intc)

	
	def close(self):
		"""
		Close object and clear memory
		"""

		# Free interval_list memory
		if self.c_aiarray:
			aiarray_destroy(self.c_aiarray)
		self.c_aiarray = NULL
		
		self.is_closed = True


import numpy as np
cimport numpy as np
cimport cython
from libc.stdint cimport uint32_t, int32_t, int64_t
from libc.stdlib cimport malloc, free
from .Interval_core cimport *
from .array_query_core cimport *


cdef extern from "src/augmented_interval_array.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/overlap_index.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/aiarray_add.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/aiarray_construct.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/aiarray_index.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/aiarray_query.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/aiarray_nhits.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/aiarray_coverage.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/aiarray_wps.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/aiarray_merge.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/aiarray_filter.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/aiarray_extract.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/aiarray_ops.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/augmented_interval_array.h":

	# C aiarray struct
	ctypedef struct aiarray_t:
		int64_t nr, mr						# Number of intervals
		interval_t *interval_list			# List of interval_t objects
		int nc								# Number of components
		int lenC[10]						# Length of components
		int idxC[10]						# Index of components
		uint32_t *maxE						# Augmentation
		uint32_t first, last				# Record range of intervals
		uint32_t *id_index					# Record index values

	# C overlap_index struct
	ctypedef struct overlap_index_t:
		int size							# Current size
		int max_size						# Maximum size
		aiarray_t *ail						# Store ailist
		long *indices						# Store indices


	#-------------------------------------------------------------------------------------
	# augmented_interval_list.c
	#=====================================================================================

	# Initialize aiarray_t
	aiarray_t *aiarray_init() nogil

	# Free aiarray data
	void aiarray_destroy(aiarray_t *ail) nogil

	# Print aiarray
	void display_array(aiarray_t *ail) nogil

	# Calculate maximum length
	int aiarray_max_length(aiarray_t *ail) nogil

	# Calculate length distribution
	void aiarray_length_distribution(aiarray_t *ail, int distribution[]) nogil


	#-------------------------------------------------------------------------------------
	# overlap_index.c
	#=====================================================================================

	# Initialize overlap_index_t
	overlap_index_t *overlap_index_init() nogil

	# Free overlap_index memory
	void overlap_index_destroy(overlap_index_t *oi) nogil

	# Add interval and index to overlap_index
	void overlap_index_add(overlap_index_t *aq, interval_t *i) nogil


	#-------------------------------------------------------------------------------------
	# aiarray_add.c
	#=====================================================================================

	# Build aiarray from arrays
	void aiarray_from_array(aiarray_t *ail, const long starts[], const long ends[], int length) nogil

	# Add a interval_t interval
	void aiarray_add(aiarray_t *ail, uint32_t start, uint32_t end) nogil

	# Append two aiarray
	aiarray_t *aiarray_append(aiarray_t *ail1, aiarray_t *ail2) nogil


	#-------------------------------------------------------------------------------------
	# aiarray_construct.c
	#=====================================================================================

	# Construct aiarray: decomposition and augmentation
	void aiarray_construct(aiarray_t *ail, int cLen) nogil


	#-------------------------------------------------------------------------------------
	# aiarray_index.c
	#=====================================================================================

	# Record id positions by index
	void aiarray_cache_id(aiarray_t *ail) nogil

	# Get interval with id
	interval_t *aiarray_get_index(aiarray_t *ail, int index) nogil

	# Get intervals with ids
	aiarray_t *aiarray_get_index_array(aiarray_t *ail, const long indices[], int length) nogil

	# Index aiarray by array
	aiarray_t *aiarray_array_index(aiarray_t *ail, const long indices[], int length) nogil

	# Index aiarray by another aiarray
	aiarray_t *aiarray_index_by_aiarray(aiarray_t *ail1, aiarray_t *ail2) nogil

	# Index aiarray by another aiarray inplace
	int aiarray_index_by_aiarray_inplace(aiarray_t *ail1, aiarray_t *ail2) nogil

	#-------------------------------------------------------------------------------------
	# aiarray_query.c
	#=====================================================================================

	# Binary search
	uint32_t binary_search(interval_t* As, uint32_t idxS, uint32_t idxE, uint32_t qe) nogil

	# Query aiarray intervals
	aiarray_t *aiarray_query(aiarray_t *ail, uint32_t qs, uint32_t qe) nogil

	# Query aiarray intervals of a length
	aiarray_t *aiarray_query_length(aiarray_t *ail, uint32_t qs, uint32_t qe, int min_length, int max_length) nogil

	# Query aiarray intervals from arrays
	array_query_t *aiarray_query_from_array(aiarray_t *ail, const long starts[], const long ends[], int length) nogil

	# Find overlaps from aiarray
	array_query_t *aiarray_query_from_aiarray(aiarray_t *ail, aiarray_t *ail2) nogil

	# Query aiarray intervals and record original index
	overlap_index_t *aiarray_query_with_index(aiarray_t *ail, uint32_t qs, uint32_t qe) nogil


	#-------------------------------------------------------------------------------------
	# aiarray_nhits.c
	#=====================================================================================

	# Determine number of hits for each query interval
	void aiarray_nhits_from_array(aiarray_t *ail, const long starts[], const long ends[],
								int length, int nhits[]) nogil

	# Determine number of hits of a length for each query interval
	void aiarray_nhits_from_array_length(aiarray_t *ail, const long starts[], const long ends[],
										int length, int nhits[], int min_length,
										int max_length) nogil

	# Calculate n hits within bins
	void aiarray_bin_nhits(aiarray_t *ail, double coverage[], int bin_size) nogil

	# Calculate n hits of a length within bins
	void aiarray_bin_nhits_length(aiarray_t *ail, double coverage[], int bin_size, int min_length, int max_length) nogil


	#-------------------------------------------------------------------------------------
	# aiarray_coverage.c
	#=====================================================================================

	# Determine coverage for an interval
	void aiarray_interval_coverage(aiarray_t *ail, int start, int end, int coverage[]) nogil

	# Calculate coverage
	void aiarray_coverage(aiarray_t *ail, double coverage[]) nogil

	# Calculate coverage within bins
	void aiarray_bin_coverage(aiarray_t *ail, double coverage[], int bin_size) nogil

	# Calculate coverage within bins of a length
	void aiarray_bin_coverage_length(aiarray_t *ail, double coverage[], int bin_size, int min_length, int max_length) nogil


	#-------------------------------------------------------------------------------------
	# aiarray_wps.c
	#=====================================================================================

	# Calculate Window Protection Score
	void aiarray_wps(aiarray_t *ail, double wps[], uint32_t protection) nogil

	# Calculate Window Protection Score of a length
	void aiarray_wps_length(aiarray_t *ail, double wps[], uint32_t protection, int min_length, int max_length) nogil


	#-------------------------------------------------------------------------------------
	# aiarray_merge.c
	#=====================================================================================

	# Get component index
	int *get_comp_bounds(aiarray_t *ail) nogil

	# Merge nearby intervals
	aiarray_t *aiarray_merge(aiarray_t *ail, uint32_t gap) nogil


	#-------------------------------------------------------------------------------------
	# aiarray_filter.c
	#=====================================================================================

	# Filter aiarray by length
	aiarray_t *aiarray_length_filter(aiarray_t *ail, int min_length, int max_length) nogil

	# Randomly downsample
	aiarray_t *aiarray_downsample(aiarray_t *ail, double proportion) nogil

	# Randomly downsample with original index
	overlap_index_t *aiarray_downsample_with_index(aiarray_t *ail, double proportion) nogil


	#-------------------------------------------------------------------------------------
	# aiarray_extract.c
	#=====================================================================================

	# Extract start for aiarray
	void aiarray_extract_starts(aiarray_t *ail, long starts[]) nogil

	# Extract end for aiarray
	void aiarray_extract_ends(aiarray_t *ail, long ends[]) nogil

	# Extract id for aiarray
	void aiarray_extract_ids(aiarray_t *ail, long ids[]) nogil


	#-------------------------------------------------------------------------------------
	# aiarray_ops.c
	#=====================================================================================

	# Subtract intervals from region
	void aiarray_subtract_intervals(aiarray_t *ref_ail, aiarray_t *result_ail, interval_t query_i, int j) nogil

	# Subtract two aiarray_t intervals
	aiarray_t *aiarray_subtract(aiarray_t *ref_ail, aiarray_t *query_ail) nogil

	# Subtract intervals from region
	void aiarray_common_intervals(aiarray_t *ref_ail, aiarray_t *result_ail, interval_t query_i, int j) nogil

	# Subtract two aiarray_t intervals
	aiarray_t *aiarray_common(aiarray_t *ref_ail, aiarray_t *query_ail) nogil


cpdef object rebuild_IntervalArray(bytes data, bytes b_length)

cdef class IntervalArray(object):
	"""
	Wrapper for C aiarray_t
	"""

	# Object attributes
	cdef aiarray_t *c_aiarray
	cdef public bint is_constructed
	cdef public bint is_closed
	cdef public bint is_frozen

	# Methods for serialization
	cdef bytes _get_data(self)
	cdef aiarray_t *_set_data(self, bytes data, bytes b_length)

	# Methods to implement C functions
	cdef void set_list(IntervalArray self, aiarray_t *input_list)
	cdef void _insert(IntervalArray self, int start, int end)
	cdef void _construct(IntervalArray self, int min_length)
	cdef aiarray_t *_array_index(IntervalArray self, const long[::1] ids)
	cdef np.ndarray _get_comp_bounds(IntervalArray self)
	cdef aiarray_t *_intersect(IntervalArray self, int start, int end)
	cpdef _intersect_from_array(IntervalArray self, const long[::1] starts, const long[::1] ends)
	cpdef _intersect_from_aiarray(IntervalArray self, IntervalArray ail)
	cpdef _intersect_with_index(IntervalArray self, int start, int end)
	cdef np.ndarray _coverage(IntervalArray self)
	cdef np.ndarray _bin_coverage(IntervalArray self, int bin_size)
	cdef np.ndarray _bin_coverage_length(IntervalArray self, int bin_size, int min_length, int max_length)
	cdef np.ndarray _bin_nhits(IntervalArray self, int bin_size)
	cdef np.ndarray _bin_nhits_length(IntervalArray self, int bin_size, int min_length, int max_length)
	cdef np.ndarray _wps(IntervalArray self, int protection)
	cdef np.ndarray _wps_length(IntervalArray self, int protection, int min_length, int max_length)
	cdef np.ndarray _length_dist(IntervalArray self)
	cdef np.ndarray _nhits_from_array(IntervalArray self, const long[::1] starts, const long[::1] ends)
	cdef np.ndarray _nhits_from_array_length(IntervalArray self, const long[::1] starts, const long[::1] ends, int min_length, int max_length)
	cdef np.ndarray _interval_coverage(IntervalArray self, int start, int end)
	cpdef _downsample_with_index(IntervalArray self, double proportion)


import numpy as np
cimport numpy as np
cimport cython
from libc.stdint cimport uint32_t, int32_t, int64_t
from libc.stdlib cimport malloc, free


cdef extern from "src/array_query_utilities.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/array_query_utilities.h":
	# C is include here so that it doesn't need to be compiled externally
	ctypedef struct array_query_t:
		long *ref_index
		long *query_index
		int size
		int max_size
	
	#-------------------------------------------------------------------------------------
	# array_query_utilities.c
	#=====================================================================================

	# Initialize array_query struct
	array_query_t *array_query_init() nogil

	# Free array_query struct memory
	void array_query_destroy(array_query_t *aq) nogil

	# Add query to array_query struct
	void array_query_add(array_query_t *aq, long ref, long query) nogil


# Convert c pointer to numpy array
cdef np.ndarray pointer_to_numpy_array(void *ptr, np.npy_intp size)


#cython: embedsignature=True
#cython: profile=False

import os
import sys
import numpy as np
cimport numpy as np
import math
cimport cython
import pandas as pd
from libc.string cimport memcpy
np.import_array()
from time import time

# Set byteorder for __reduce__
byteorder = sys.byteorder


def get_include():
	"""
	Get file directory if C headers

	Parameters
	----------
		None

	Returns
	-------
		location : str
			Directory to header files
	"""

	# Grab file location
	location = os.path.split(os.path.realpath(__file__))[0]

	return location


cdef np.ndarray pointer_to_numpy_array(void *ptr, np.npy_intp size):
	"""
	Convert c pointer to numpy array.
	The memory will be freed as soon as the ndarray is deallocated.

	Parameters
	----------
		ptr : void
			Pointer to be given to numpy
		size : np.npy_intp
			Size of the array

	Returns
	-------
		arr : numpy.ndarray
			Numpy array from given pointer

	"""

	# Import functions for numpy C header
	cdef extern from "numpy/arrayobject.h":
		void PyArray_ENABLEFLAGS(np.ndarray arr, int flags)

	# Create shape of ndarray
	cdef np.npy_intp dims[1]
	dims[0] = size

	# Create ndarray from C pointer
	cdef np.ndarray arr = np.PyArray_SimpleNewFromData(1, &dims[0], np.NPY_LONG, ptr)
	
	# Hand control of data freeing to numpy
	PyArray_ENABLEFLAGS(arr, np.NPY_ARRAY_OWNDATA)
	#np.PyArray_UpdateFlags(arr, arr.flags.num | np.NPY_OWNDATA)

	return arr



from __future__ import absolute_import
from .Interval_core import Interval, get_include
from .LabeledInterval_core import LabeledInterval
from .AIList_core import AIList
from .IntervalArray_core import IntervalArray
from .LabeledIntervalArray_core import LabeledIntervalArray

# This is extracted automatically by the top-level setup.py.
__version__ = '1.0.2'
__author__ = "Kyle S. Smith"


__doc__ = """\
API
======

Basic class
-----------

.. autosummary::
   :toctree: .
   
   Interval
   LabeledInterval
   AIList
   IntervalArray
   LabeledIntervalArray
    
"""


import numpy as np
cimport numpy as np
cimport cython
from libc.stdint cimport uint32_t, int32_t, int64_t, uint16_t
from libc.stdlib cimport malloc, free
from .LabeledInterval_core cimport *
from .array_query_core cimport *
ctypedef np.uint8_t uint8


cdef extern from "src/labeled_augmented_array.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/overlap_label_index.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/labeled_aiarray_add.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/labeled_aiarray_get.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/labeled_aiarray_construct.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/labeled_aiarray_index.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/labeled_aiarray_query.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/labeled_aiarray_has_hit.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/labeled_aiarray_nhits.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/labeled_aiarray_wps.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/labeled_aiarray_coverage.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/labeled_aiarray_merge.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/labeled_aiarray_filter.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/labeled_aiarray_extract.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/labeled_aiarray_iter.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/labeled_aiarray_match.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/khash.h":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/labeled_augmented_array.h":

	ctypedef struct labeled_aiarray_t:
		int64_t nr							# Number of intervals
		int64_t mr							# Max number of intervals
		int64_t nl							# Number of unique labels
		int64_t ml							# Max number of unique labels
		labeled_interval_t *interval_list	# List of interval_t objects
		int *nc								# Number of components
		int *lenC							# Length of components
		int *idxC							# Index of components in interval_list
		uint32_t *maxE						# Augmentation
		uint32_t *first						# Record first position of intervals
		uint32_t *last						# Record last position of intervals
		uint32_t *label_count				# Record number of label values
		void *label_map						# Hash table to convert labels to int
		void *rev_label_map					# Hash table to convert labels to int
	
	# C overlap_label_index struct
	ctypedef struct overlap_label_index_t:
		int size							# Current size
		int max_size						# Maximum size
		labeled_aiarray_t *ail				# Store labeled_aiarray
		long *indices						# Store indices

	ctypedef struct label_sorted_iter_t:
		labeled_aiarray_t *ail				# Labeled array
		uint16_t label						# Label
		int nc								# Number of components
		int label_start						# First index for label
		int label_end						# Last index for label
		int *label_comp_bounds				# Label component bounds
		int *label_comp_used				# Components used
		labeled_interval_t *intv			# Interval
		int n								# Current position


	#-------------------------------------------------------------------------------------
	# labeled_augmented_array.c
	#=====================================================================================

	# Initialize aiarray_t
	labeled_aiarray_t *labeled_aiarray_init() nogil

	# Free aiarray data
	void labeled_aiarray_destroy(labeled_aiarray_t *ail) nogil

	# Print aiarray
	void labeled_display_list(labeled_aiarray_t *ail) nogil

	# Print label map
	void display_label_map(labeled_aiarray_t *ail) nogil

	# Calculate maximum length
	int labeled_aiarray_max_length(labeled_aiarray_t *ail) nogil

	# Calculate length distribution
	void labeled_aiarray_length_distribution(labeled_aiarray_t *ail, int distribution[]) nogil


	#-------------------------------------------------------------------------------------
	# overlap_label_index.c
	#=====================================================================================

	# Initialize overlap_label_index_t
	overlap_label_index_t *overlap_label_index_init() nogil

	# Free overlap_label_index memory
	void overlap_label_index_destroy(overlap_label_index_t *oi) nogil

	# Add interval and index to overlap_label_index
	void overlap_label_index_add(overlap_label_index_t *aq, labeled_interval_t i, const char *label_name) nogil


	#-------------------------------------------------------------------------------------
	# labeled_aiarray_add.c
	#=====================================================================================

	# Add a interval_t from array
	void labeled_aiarray_from_array(labeled_aiarray_t *ail, const long starts[], const long ends[], const char label_names[], int length, int label_str_len) nogil

	# Add a interval_t interval
	void labeled_aiarray_add(labeled_aiarray_t *ail, uint32_t start, uint32_t end, const char *label_name) nogil

	# Add intervals from another labeled_aiarray
	void labeled_aiarray_append(labeled_aiarray_t *ail, labeled_aiarray_t *ail2) nogil

	# Copy labeled_aiarray
	labeled_aiarray_t *labeled_aiarray_copy(labeled_aiarray_t *ail) nogil


	#-------------------------------------------------------------------------------------
	# labeled_aiarray_get.c
	#=====================================================================================

	# Determine if label_name is in array
	int label_is_present(labeled_aiarray_t *ail, const char *label_name) nogil

	# Query label map for label name
	uint16_t query_label_map(labeled_aiarray_t *ail, const char *label_name) nogil

	# Query rev_label_map for label
	const char *query_rev_label_map(labeled_aiarray_t *ail, uint16_t label) nogil

	# Get intervals with label name
	labeled_aiarray_t *get_label(labeled_aiarray_t *ail, const char *label_name) nogil

	# Get intervals with label name and original index
	overlap_label_index_t *get_label_with_index(labeled_aiarray_t *ail, const char *label_name) nogil

	# Get intervals with labels names from array
	labeled_aiarray_t *get_label_array(labeled_aiarray_t *ail, const char *label_names[], int length) nogil

	# Get intervals with labels from array and original index
	overlap_label_index_t *get_label_array_with_index(labeled_aiarray_t *ail, const char label_names[], int n_labels, int label_str_len) nogil

	# Get index start of label
	int get_label_index(labeled_aiarray_t *ail, int label) nogil

	# Determine label indices
	void get_label_index_array(labeled_aiarray_t *ail, int *label_index) nogil

	# Get ids for labels in a label array
	void get_label_array_ids(labeled_aiarray_t *ail, const char *label_names[], int n_labels, long index[]) nogil

	# Determine if an index is of a label array
	void get_label_array_presence(labeled_aiarray_t *ail, const char label_names[], int n_labels, uint8 index[], int label_str_len) nogil


	#-------------------------------------------------------------------------------------
	# labeled_aiarray_construct.c
	#=====================================================================================

	# Sort intervals in aiarray
	void labeled_aiarray_sort(labeled_aiarray_t *ail) nogil

	# Sort intervals by label
	void labeled_aiarray_radix_label_sort(labeled_aiarray_t *ail) nogil

	# Construct aiarray: decomposition and augmentation
	void labeled_aiarray_construct(labeled_aiarray_t *ail, int cLen_init) nogil


	#-------------------------------------------------------------------------------------
	# labeled_aiarray_iter.c
	#=====================================================================================

	# Get component index for label
	int *get_label_comp_bounds(labeled_aiarray_t *ail, int label) nogil

	label_sorted_iter_t *iter_init(labeled_aiarray_t *ail, const char *label_name) nogil

	int iter_next(label_sorted_iter_t *iterator) nogil

	void iter_destroy(label_sorted_iter_t *iterator) nogil


	#-------------------------------------------------------------------------------------
	# labeled_aiarray_match.c
	#=====================================================================================

	overlap_label_index_t *has_exact_match(labeled_aiarray_t *ail1, labeled_aiarray_t *ail2) nogil

	#-------------------------------------------------------------------------------------
	# labeled_aiarray_index.c
	#=====================================================================================

	# Record id positions by id
	void labeled_aiarray_cache_id(labeled_aiarray_t *ail) nogil

	# Get intervals with ids
	labeled_aiarray_t *labeled_aiarray_slice_index(labeled_aiarray_t *ail, const long ids[], int length) nogil

	# Get intervals with range
	labeled_aiarray_t *labeled_aiarray_slice_range(labeled_aiarray_t *ail, int start, int end, int step) nogil

	# Get intervals with boolean array
	labeled_aiarray_t *labeled_aiarray_slice_bool(labeled_aiarray_t *ail, uint8 bool_index[]) nogil

	# Get interval with id
	labeled_interval_t *labeled_aiarray_get_id(labeled_aiarray_t *ail, int id_value) nogil

	# Index aiarray by another aiarray inplace
	int labeled_aiarray_index_by_aiarray_inplace(labeled_aiarray_t *ail1, labeled_aiarray_t *ail2) nogil


	#-------------------------------------------------------------------------------------
	# labeled_aiarray_query.c
	#=====================================================================================

	# Binary search
	uint32_t binary_search_labeled(labeled_interval_t* As, uint32_t idxS, uint32_t idxE, uint32_t qe) nogil

	# Base query logic
	void labeled_aiarray_query(labeled_aiarray_t *ail, labeled_aiarray_t *overlaps, uint32_t qs, uint32_t qe, const char *label_name) nogil

	# Base query logic filtered by length
	void labeled_aiarray_query_length(labeled_aiarray_t *ail, labeled_aiarray_t *overlaps, uint32_t qs, uint32_t qe, const char *label_name, int min_length, int max_length) nogil

	# Base query logic for nhits
	void labeled_aiarray_query_nhits(labeled_aiarray_t *ail, long *nhits, uint32_t qs, uint32_t qe, const char *label_name) nogil

	# Base query logic for nhits filtered by length
	void labeled_aiarray_query_nhits_length(labeled_aiarray_t *ail, long *nhits, uint32_t qs, uint32_t qe, const char *label_name, int min_length, int max_length) nogil

	# Base query logic with index
	void labeled_aiarray_query_with_index(labeled_aiarray_t *ail, overlap_label_index_t *overlaps, uint32_t qs, uint32_t qe, const char *label_name) nogil

	# Base query logic with index
	void labeled_aiarray_query_only_index(labeled_aiarray_t *ail, array_query_t *overlaps, uint32_t qs, uint32_t qe, uint32_t id_value, const char *label_name) nogil

	# Base query logic if present
	void labeled_aiarray_query_has_hit(labeled_aiarray_t *ail, uint8 has_hit[], uint32_t qs, uint32_t qe, const char *label_name) nogil

	# Query interval
	labeled_aiarray_t *labeled_aiarray_query_single(labeled_aiarray_t *ail, uint32_t qs, uint32_t qe, const char *label_name) nogil

	# Query aiarray intervals of a length
	labeled_aiarray_t *labeled_aiarray_query_single_length(labeled_aiarray_t *ail, uint32_t qs, uint32_t qe, const char *label_name, int min_length, int max_length) nogil

	# Query aiarray intervals from arrays
	array_query_t *labeled_aiarray_query_from_array(labeled_aiarray_t *ail, const long starts[], const long ends[], const char label_names[], int length, int label_str_len) nogil

	# Query aiarray intervals from aiarray
	array_query_t *labeled_aiarray_query_from_labeled_aiarray(labeled_aiarray_t *ail, labeled_aiarray_t *ail2) nogil

	# Query aiarray intervals and record original index
	overlap_label_index_t *labeled_aiarray_query_single_with_index(labeled_aiarray_t *ail, uint32_t qs, uint32_t qe, const char *label_name) nogil

	# Query aiarray intervals and record original index from labled_aiarray
	overlap_label_index_t *labeled_aiarray_query_with_index_from_labeled_aiarray(labeled_aiarray_t *ail, labeled_aiarray_t *ail2) nogil

	
	#-------------------------------------------------------------------------------------
	# labeled_aiarray_has_hits.c
	#=====================================================================================

	# Determine if hit is present for each query interval
	void labeled_aiarray_has_hit_from_array(labeled_aiarray_t *ail, const long starts[], const long ends[],
                                        const char label_names[], int length, int label_str_len, uint8 has_hit[]) nogil

	#-------------------------------------------------------------------------------------
	# labeled_aiarray_nhits.c
	#=====================================================================================

	# Determine number of hits for interval
	long labeled_aiarray_nhits(labeled_aiarray_t *ail, long start, long end, const char *label_name) nogil

	# Determine number of hits for interval of a length
	long labeled_aiarray_nhits_length(labeled_aiarray_t *ail, long start, long end, const char *label_name, int min_length, int max_length) nogil

	# Determine number of hits for each query interval
	void labeled_aiarray_nhits_from_array(labeled_aiarray_t *ail, const long starts[], const long ends[],
											const char *label_names[], int length, int nhits[]) nogil

	# Determine number of hits of a length for each query interval
	void labeled_aiarray_nhits_from_array_length(labeled_aiarray_t *ail, const long starts[], const long ends[],
												const char *label_names[], int length, int nhits[], int min_length,
												int max_length) nogil

	# Calculate n hits within bins
	void labeled_aiarray_bin_nhits(labeled_aiarray_t *ail, labeled_aiarray_t *bins, double nhits[], int bin_size) nogil

	# Calculate n hits of a length within bins
	void labeled_aiarray_bin_nhits_length(labeled_aiarray_t *ail, labeled_aiarray_t *bins, double nhits[], int bin_size,
											int min_length, int max_length) nogil


	#-------------------------------------------------------------------------------------
	# labeled_aiarray_wps.c
	#=====================================================================================

	# Calculate Window Protection Score for a label
	void labeled_aiarray_label_wps(labeled_interval_t *interval_list, double wps[], uint32_t protection, int nr, int first, int last) nogil

	# Calculate Window Protection Score
	void labeled_aiarray_wps(labeled_aiarray_t *ail, double wps[], uint32_t protection) nogil

	# Calculate Window Protection Score for a label of a length
	void labeled_aiarray_label_wps_length(labeled_interval_t *interval_list, double wps[], uint32_t protection, int nr, int first, int last,
											int min_length, int max_length) nogil

	# Calculate Window Protection Score of a length
	void labeled_aiarray_wps_length(labeled_aiarray_t *ail, double wps[], uint32_t protection, int min_length, int max_length) nogil


	#-------------------------------------------------------------------------------------
	# labeled_aiarray_coverage.c
	#=====================================================================================

	# Determine coverage for an interval
	void labeled_aiarray_interval_coverage(labeled_aiarray_t *ail, int start, int end, const char *label_name, int coverage[]) nogil

	# Calculate coverage for a label
	void labeled_aiarray_label_coverage(labeled_interval_t *interval_list, double coverage[], const char *label_name, int nr) nogil

	# Calculate coverage
	void labeled_aiarray_coverage(labeled_aiarray_t *ail, double coverage[]) nogil

	#-------------------------------------------------------------------------------------
	# labeled_aiarray_merge.c
	#=====================================================================================

	# Merge nearby intervals
	labeled_aiarray_t *labeled_aiarray_merge(labeled_aiarray_t *ail, uint32_t gap) nogil


	#-------------------------------------------------------------------------------------
	# labeled_aiarray_filter.c
	#=====================================================================================

	# Filter labeled_aiarray by length
	labeled_aiarray_t *labeled_aiarray_length_filter(labeled_aiarray_t *ail, int min_length, int max_length) nogil

	# Randomly downsample
	labeled_aiarray_t *labeled_aiarray_downsample(labeled_aiarray_t *ail, double proportion) nogil

	# Randomly downsample with original index
	overlap_label_index_t *labeled_aiarray_downsample_with_index(labeled_aiarray_t *ail, double proportion) nogil


	#-------------------------------------------------------------------------------------
	# labeled_aiarray_extract.c
	#=====================================================================================

	# Extract start for labeled_aiarray
	void labeled_aiarray_extract_starts(labeled_aiarray_t *ail, long starts[]) nogil

	# Extract end for labeled_aiarray
	void labeled_aiarray_extract_ends(labeled_aiarray_t *ail, long ends[]) nogil

	# Extract id for labeled_aiarray
	void labeled_aiarray_extract_ids(labeled_aiarray_t *ail, long ids[]) nogil


cpdef object rebuild_LabeledIntervalArray(bytes data, bytes b_length, bytes b_label_list)

cdef class LabeledIntervalArray(object):
	"""
	Wrapper for C aiarray_t
	"""

	# Define class properties
	cdef labeled_aiarray_t *ail
	cdef public bint is_constructed
	cdef public bint is_closed
	cdef public bint is_frozen

	# Functions for serialization
	cdef bytes _get_data(self)
	cdef labeled_aiarray_t *_set_data(self, bytes data, bytes b_length, bytes b_label_list)

	# Class methods
	cdef void set_list(LabeledIntervalArray self, labeled_aiarray_t *input_list)

	cdef labeled_aiarray_t *_array_index(LabeledIntervalArray self, const long[::1] ids)
	cdef np.ndarray _get_index(LabeledIntervalArray self, str label)
	cdef np.ndarray _get_index_multi(LabeledIntervalArray self, np.ndarray labels)

	cdef void _insert(LabeledIntervalArray self, int start, int end, const char *label)
	cdef void _construct(LabeledIntervalArray self, int min_length)
	cdef void _from_array(LabeledIntervalArray self, const long[::1] starts, const long[::1] ends, const char *label_names, int array_length, int label_str_len)
	cdef void _append(LabeledIntervalArray self, LabeledIntervalArray other_ail)

	cdef int _get_label_index(LabeledIntervalArray self, int label)
	cdef np.ndarray _get_label_comp_bounds(LabeledIntervalArray self, int label)
	cdef np.ndarray _get_label_comp_length(LabeledIntervalArray self, int label)
	cdef labeled_aiarray_t *get_labels(LabeledIntervalArray self, const char[:,::1] labels, int length)
	cdef overlap_label_index_t *get_labels_with_index(LabeledIntervalArray self, const char *label_names, int str_label_len, int length)

	cdef labeled_aiarray_t *_intersect(LabeledIntervalArray self, int start, int end, const char *label_name)
	cpdef _intersect_from_array(LabeledIntervalArray self, const long[::1] starts, const long[::1] ends, const char *labels, int label_str_len)
	cpdef _has_hit_from_array(LabeledIntervalArray self, const long[::1] starts, const long[::1] ends, uint8[::1] has_hit, const char *labels, int label_str_len)
	cpdef _intersect_from_labeled_aiarray(LabeledIntervalArray self, LabeledIntervalArray ail)
	cpdef _intersect_with_index(LabeledIntervalArray self, int start, int end, const char *label_name)
	cpdef _intersect_with_index_from_LabeledIntervalArray(LabeledIntervalArray self, LabeledIntervalArray ail2)

	cdef np.ndarray _nhits_from_array(LabeledIntervalArray self, const long[::1] starts, const long[::1] ends, const char[:,::1] labels)
	cdef np.ndarray _nhits_from_array_length(LabeledIntervalArray self, const long[::1] starts, const long[::1] ends, const char[:,::1] labels, int min_length, int max_length)

	cdef labeled_aiarray_t *_determine_bins(LabeledIntervalArray self, int bin_size)
	cdef np.ndarray _bin_nhits(LabeledIntervalArray self, labeled_aiarray_t *bins, int bin_size)
	cdef np.ndarray _bin_nhits_length(LabeledIntervalArray self, labeled_aiarray_t *bins, int bin_size, int min_length, int max_length)

	cdef np.ndarray _wps(LabeledIntervalArray self, int protection)
	cdef np.ndarray _wps_length(LabeledIntervalArray self, int protection, int min_length, int max_length)

	cdef np.ndarray _coverage(LabeledIntervalArray self)

	cpdef _downsample_with_index(LabeledIntervalArray self, double proportion)
	cdef np.ndarray _length_dist(LabeledIntervalArray self)
	cpdef _filter_exact_match(LabeledIntervalArray self, LabeledIntervalArray other_aiarray)
	cdef np.ndarray _has_exact_match(LabeledIntervalArray self, LabeledIntervalArray other_aiarray)
	cdef int _index_with_aiarray(LabeledIntervalArray self, LabeledIntervalArray other_aiarray)
	cdef void _get_locs(LabeledIntervalArray self, uint8[::1] locs_view, char *label_names, int label_str_len, int n_labels)


import numpy as np
cimport numpy as np
cimport cython
from libc.stdint cimport uint32_t, int32_t, int64_t, uint16_t
from libc.stdlib cimport malloc, free


cdef extern from "src/labeled_interval.c":
	# C is include here so that it doesn't need to be compiled externally
	pass

cdef extern from "src/labeled_interval.h":
	# C is include here so that it doesn't need to be compiled externally
	ctypedef struct labeled_interval_t:
		uint32_t start      				# Region start: 0-based
		uint32_t end    					# Region end: not inclusive
		int32_t id_value					# Region ID
		uint16_t label						# Region label

	# Initialize interval_t
	labeled_interval_t *labeled_interval_init(uint32_t start, uint32_t end, int32_t id_value, uint16_t label) nogil


cdef class LabeledInterval(object):
	"""
	Wrapper for C labeled interval
	"""

	# Define attributes
	cdef str _label

	# C instance of struct
	cdef labeled_interval_t *i

	# Methods for serialization
	cdef void set_i(self, labeled_interval_t *i, str label)
